@@ -1,12 +1,8 @@
 import functools
-import inspect
 import json
-import logging
-import time
-from functools import singledispatch, update_wrapper
-from typing import Any, Callable, Dict, Optional, Tuple, Union
+from typing import Any, Callable, Dict, Iterable, Optional, Tuple, Union, overload
 
-from flask import Blueprint, Request, Response, current_app, g, jsonify, request
+from flask import Blueprint, Response, current_app, g, jsonify, request
 from jsonschema.validators import Draft7Validator
 from werkzeug.exceptions import BadRequest, NotFound, NotImplemented, Unauthorized
 
@@ -16,6 +12,7 @@ from globus_action_provider_tools.authorization import (
     authorize_action_management_or_404,
 )
 from globus_action_provider_tools.data_types import (
+    ActionProviderDescription,
     ActionProviderJsonEncoder,
     ActionRequest,
     ActionStatus,
@@ -27,67 +24,59 @@ from globus_action_provider_tools.flask import (
 )
 
 ActionStatusReturn = Union[ActionStatus, Tuple[ActionStatus, int]]
-ActionLogReturn = Dict[str, Any]
-ViewReturn = Union[Tuple[Response, int], Tuple[str, int]]
+ActionStatusType = Callable[[Union[str, ActionStatus], AuthState], ActionStatusReturn]
 
-ActionRunType = Callable[[ActionRequest, AuthState], ActionStatusReturn]
-ActionStatusType = Union[
-    Callable[[str, AuthState], ActionStatusReturn],
-    Callable[[ActionStatus, AuthState], ActionStatusReturn],
-]
-ActionCancelType = ActionStatusType
-ActionReleaseType = ActionStatusType
+ActionLogReturn = Dict[str, Any]
 ActionLogType = Callable[[str, AuthState], ActionLogReturn]
 
+ActionRunType = Callable[[ActionRequest, AuthState], ActionStatusReturn]
+ActionCancelType = ActionStatusType
+ActionReleaseType = ActionStatusType
 
-def methdispatch(func):
-    """
-    Allows singledispatch to work on a mix of Class methods and regular functions.
-    Huge shoutout to https://stackoverflow.com/a/24602374.
+ViewReturn = Union[Tuple[Response, int], Tuple[str, int]]
 
-    This decorator decorates a base function which we want to call in a
-    "generic" way where generic means that the first argument to the function
-    call can have multiple types, and the function called differs
-    according to the arg type supplied. The dispatch is essentially a lookup
-    based on type:
-
-    {
-        class:Object: my_func,
-        class:Str: my_func_for_strs,
-        class:Int: my_func_for_ints,
-        class:ActionStatus: my_func_for_ActionStatus
-    }
-    """
-    dispatcher = singledispatch(func)
-
-    def wrapper(*args, **kw):
-        matching_func = dispatcher.dispatch(args[1].__class__)
-
-        # If the dispatched function is user-supplied, it isn't an instance
-        # method so we remove the first arg which is the "self" reference
-        if matching_func is not dispatcher.dispatch(object):
-            args = args[1:]
-
-        return matching_func(*args, **kw)
-
-    wrapper.register = dispatcher.register
-    update_wrapper(wrapper, func)
-    return wrapper
+ActionLoaderType = Tuple[Callable[[str, Any], ActionStatus], Any]
+ActionSaverType = Tuple[Callable[[ActionStatus, Any], None], Any]
 
 
 class ActionProviderBlueprint(Blueprint):
     def __init__(
-        self, provider_description, *args, **kwarg,
+        self,
+        provider_description: ActionProviderDescription,
+        *args,
+        globus_auth_client_name: Optional[str] = None,
+        additional_scopes: Iterable[str] = (),
+        **kwarg,
     ):
+        """Create a new ActionProviderBlueprint. All arguments not listed here are the
+        same as a Flask Blueprint.
+
+        :param provider_description: A Provider Description which will be
+        returned from introspection calls to this Blueprint.
+
+        :param globus_auth_client_name: The name of the Globus Auth Client (also
+        known as the resource server name). This will be used to validate the
+        intended audience for tokens passed to the operations on this
+        Blueprint. By default, the client id will be used for checkign audience,
+        and unless the client has explicitly been given a resource server name
+        in Globus Auth, this will be proper behavior.
+
+        :param additional_scopes: Additional scope strings the Action Provider
+        should allow scopes in addition to the one specified by the
+        ``globus_auth_scope`` value of the input provider description. Only
+        needed if more than one scope has been allocated for the Action
+        Provider's Globus Auth client_id.
+        """
+
         super().__init__(*args, **kwarg)
 
-        self.action_status_plugin: ActionStatusType = None
-        self.action_cancel_plugin: ActionCancelType = None
-        self.action_loader_plugin = None
-        self.action_saver_plugin = None
+        self.action_loader_plugin: Optional[ActionLoaderType] = None
+        self.action_saver_plugin: Optional[ActionSaverType] = None
 
         self.provider_description = provider_description
         self.input_body_validator = self._load_input_body_validator()
+        self.globus_auth_client_name = globus_auth_client_name
+        self.additional_scopes = additional_scopes
 
         self.json_encoder = ActionProviderJsonEncoder
         self.before_request(self._check_token)
@@ -111,13 +100,13 @@ class ActionProviderBlueprint(Blueprint):
 
     def register(self, app, options, first_registration=False):
         """
-        Override the built in Blueprint register function to allow our 
+        Override the built in Blueprint register function to allow our
         Blueprint to pull configuration data once it is registered with the
         Flask app to create its internal instance of a TokenChecker.
 
         We first check the environment to see if a blueprint-specific client ID
         and client secret were provided. If we cannot pull those from the
-        environment, we backoff and search for generic client id and secret values.  
+        environment, we backoff and search for generic client id and secret values.
         """
         provider_prefix = self.name.upper() + "_"
         client_id = app.config.get(provider_prefix + "CLIENT_ID")
@@ -127,10 +116,14 @@ class ActionProviderBlueprint(Blueprint):
             client_id = app.config.get("CLIENT_ID")
             client_secret = app.config.get("CLIENT_SECRET")
 
+        scopes = [self.provider_description.globus_auth_scope]
+        scopes.extend(self.additional_scopes)
+
         self.checker = TokenChecker(
             client_id=client_id,
             client_secret=client_secret,
-            expected_scopes=[self.provider_description.globus_auth_scope],
+            expected_scopes=scopes,
+            expected_audience=self.globus_auth_client_name,
         )
         super().register(app, options, first_registration)
 
@@ -152,7 +145,7 @@ class ActionProviderBlueprint(Blueprint):
 
     def action_run(self, func: ActionRunType) -> Callable[[], ViewReturn]:
         """
-        Decorates a function to be run as an Action Provider's run endpoint. 
+        Decorates a function to be run as an Action Provider's run endpoint.
         """
 
         @functools.wraps(func)
@@ -178,28 +171,37 @@ class ActionProviderBlueprint(Blueprint):
             self._validate_input(action_request.body)
             status = func(action_request, g.auth_state)
             self._save_action(status)
-            return self._action_status_return_to_view_return(status, 201)
+            return self._action_status_return_to_view_return(status, 202)
 
         self.add_url_rule("/run", func.__name__, wrapper, methods=["POST"])
         print(f'Registered action run plugin "{func.__name__}"')
         return wrapper
 
-    @methdispatch
-    def _generic_action_status(self, *args, **kwargs):
-        raise NotImplemented(
-            f"No action_status function registered to handle type {type(args[0])}"
-        )
+    @overload
+    def action_status(self, func: Callable[[str, AuthState], ActionStatusReturn]):
+        """
+        Using these stubs w/ @overload tells mypy that the actual implementation
+        for action_status can accept a str or ActionStatus as the first arg type
+        NOTE: typing_extensions.Protocol would be better if not for it's poor
+        error messages
+        """
+        ...
 
-    def action_status(self, func: ActionStatusType):
+    @overload
+    def action_status(
+        self, func: Callable[[ActionStatus, AuthState], ActionStatusReturn]
+    ):
+        """
+        Using these stubs w/ @overload tells mypy that the actual implementation
+        for action_status can accept a str or ActionStatus as the first arg type
+        """
+        ...
+
+    def action_status(self, func) -> None:
         """
         Decorates a function to be run as an Action Provider's status endpoint.
         """
-        _, first_param = list(inspect.signature(func).parameters.items())[0]
-        if first_param.annotation == first_param.empty:
-            # There's no annotation available, assume the function handles str
-            self._generic_action_status.register(str, func)
-        else:
-            self._generic_action_status.register(first_param.annotation, func)
+        self._action_status: ActionStatusType = func
         print(f'Registered action status plugin "{func.__name__}"')
 
     def _auto_action_status(self, action_id: str) -> ViewReturn:
@@ -214,31 +216,42 @@ class ActionProviderBlueprint(Blueprint):
             action = self._load_action_by_id(action_id)
             authorize_action_access_or_404(action, g.auth_state)
             try:
-                action = self._generic_action_status(action, g.auth_state)
-            except NotImplemented:
+                result = self._action_status(action, g.auth_state)
+            except AttributeError:
                 # Once an action_loader is registered, there is no reason to
-                # register an action_status. Therefore, an exception here is ok
-                pass
+                # register an action_status. Therefore, an exception here is ok,
+                # in which case just return the bare action as the result
+                result = action
         else:
-            action = self._generic_action_status(action_id, g.auth_state)
-        return self._action_status_return_to_view_return(action, 200)
+            try:
+                result = self._action_status(action_id, g.auth_state)
+            except AttributeError:
+                raise NotImplemented("No status endpoint is available")
+        return self._action_status_return_to_view_return(result, 200)
 
-    @methdispatch
-    def _generic_action_cancel(self, *args, **kwargs):
-        raise NotImplemented(
-            f"No action_cancel function registered to handle type {type(args[0])}"
-        )
+    @overload
+    def action_cancel(self, func: Callable[[str, AuthState], ActionStatusReturn]):
+        """
+        Using these stubs w/ @overload tells mypy that the actual implementation
+        for action_cancel can accept a str or ActionStatus as the first arg type
+        """
+        ...
 
-    def action_cancel(self, func: ActionCancelType) -> None:
+    @overload
+    def action_cancel(
+        self, func: Callable[[ActionStatus, AuthState], ActionStatusReturn]
+    ):
+        """
+        Using these stubs w/ @overload tells mypy that the actual implementation
+        for action_cancel can accept a str or ActionStatus as the first arg type
+        """
+        ...
+
+    def action_cancel(self, func) -> None:
         """
         Decorates a function to be run as an Action Provider's cancel endpoint.
         """
-        _, first_param = list(inspect.signature(func).parameters.items())[0]
-        if first_param.annotation == first_param.empty:
-            # There's no annotation available, assume the function handles str
-            self._generic_action_cancel.register(str, func)
-        else:
-            self._generic_action_cancel.register(first_param.annotation, func)
+        self._action_cancel: ActionCancelType = func
         print(f'Registered action cancel plugin "{func.__name__}"')
 
     def _auto_action_cancel(self, action_id: str) -> ViewReturn:
@@ -251,19 +264,41 @@ class ActionProviderBlueprint(Blueprint):
             action = self._load_action_by_id(action_id)
             authorize_action_management_or_404(action, g.auth_state)
             try:
-                action = self._generic_action_cancel(action, g.auth_state)
-            except NotImplemented as e:
+                result = self._action_cancel(action, g.auth_state)
+            except AttributeError:
                 # Once an action_loader is registered, if the ActionProvider is
                 # synchronous, there is no reason to register an action_cancel.
-                # Therefore, an exception here might be ok
-                pass
+                # Therefore, an exception here might be ok, in which case just
+                # return the bare action as the result
+                result = action
             finally:
                 self._save_action(action)
         else:
-            action = self._generic_action_cancel(action_id, g.auth_state)
-        return self._action_status_return_to_view_return(action, 200)
+            try:
+                result = self._action_cancel(action_id, g.auth_state)
+            except AttributeError:
+                raise NotImplemented("No cancel endpoint is available")
+        return self._action_status_return_to_view_return(result, 200)
 
-    def action_release(self, func: ActionReleaseType) -> Callable[[str], ViewReturn]:
+    @overload
+    def action_release(self, func: Callable[[str, AuthState], ActionStatusReturn]):
+        """
+        Using these stubs w/ @overload tells mypy that the actual implementation
+        for action_release can accept a str or ActionStatus as the first arg type
+        """
+        ...
+
+    @overload
+    def action_release(
+        self, func: Callable[[ActionStatus, AuthState], ActionStatusReturn]
+    ):
+        """
+        Using these stubs w/ @overload tells mypy that the actual implementation
+        for action_release can accept a str or ActionStatus as the first arg type
+        """
+        ...
+
+    def action_release(self, func) -> Callable[[str], ViewReturn]:
         """
         Decorates a function to be run as an Action Provider's release endpoint.
         """
@@ -277,10 +312,13 @@ class ActionProviderBlueprint(Blueprint):
                 authorize_action_management_or_404(action, g.auth_state)
 
             status = func(action_id, g.auth_state)  # type: ignore
-            return jsonify(status), 200
+            return self._action_status_return_to_view_return(status, 200)
 
         self.add_url_rule(
-            "/<string:action_id>/release", func.__name__, wrapper, methods=["POST"],
+            "/<string:action_id>/release",
+            func.__name__,
+            wrapper,
+            methods=["POST"],
         )
         print(f'Registered action release plugin "{func.__name__}"')
         return wrapper
@@ -322,8 +360,8 @@ class ActionProviderBlueprint(Blueprint):
 
         # TODO figure out how to get a type annotation working on this inner func
         def wrapper(func):
-            print(f"Registered action loader '{func.__name__}'")
             self.action_loader_plugin = (func, storage_backend)
+            print(f"Registered action loader '{func.__name__}'")
 
         return wrapper
 
@@ -332,6 +370,9 @@ class ActionProviderBlueprint(Blueprint):
         If the actiond_id is not found in the registered action_loader, we
         assume the ActionStatus is non-recoverable.
         """
+        if self.action_loader_plugin is None:
+            raise NotFound
+
         func, backend = self.action_loader_plugin
         action = func(action_id, backend)
         if action:
@@ -345,17 +386,17 @@ class ActionProviderBlueprint(Blueprint):
         action_savers with different backends can be registered and each backend
         will save a copy of the ActionStatus. If any action_savers are
         registered and they fail to save an ActionStatus, an error will be
-        thrown. 
+        thrown.
 
         #TODO pickup here, what should the behavior for an action_saver be? When
         should it be called?
-        # TODO the return of the action_loader is an action, the return for the plugins 
+        # TODO the return of the action_loader is an action, the return for the plugins
         # are actionStatusReturns, which are DIFFERENT [THANKS MYPY]
         """
 
         def wrapper(func):
+            self.action_saver_plugin: ActionSaverType = (func, storage_backend)
             print(f"Registered action saver '{func.__name__}'")
-            self.action_saver_plugin = (func, storage_backend)
 
         return wrapper
 
@@ -376,7 +417,7 @@ class ActionProviderBlueprint(Blueprint):
     def _check_token(self) -> None:
         """
         Parses a token from a request to generate an auth_state object which is
-        then made available as the second argument to decorated functions. 
+        then made available as the second argument to decorated functions.
         """
         access_token = (
             request.headers.get("Authorization", "").strip().lstrip("Bearer ")
@@ -439,6 +480,6 @@ class ActionProviderBlueprint(Blueprint):
         """
         if isinstance(status, ActionStatus):
             status_code = default_status_code
-        else:
+        elif isinstance(status, tuple):
             status, status_code = status
         return jsonify(status), status_code
